from django.contrib.admin import ModelAdmin
from django.contrib.admin.decorators import register
from edc_registration.admin import RegisteredSubjectAdmin as BaseRegisteredSubjectAdmin

from ..admin_site import edc_pharmacy_admin
from ..models import Subject, VisitSchedule


@register(Subject, site=edc_pharmacy_admin)
class SubjectAdmin(BaseRegisteredSubjectAdmin):
    ordering = ("subject_identifier",)
    search_fields = ("subject_identifier",)


@register(VisitSchedule, site=edc_pharmacy_admin)
class VisitScheduleAdmin(ModelAdmin):
    ordering = ("visit_schedule_name", "schedule_name", "visit_code")
    search_fields = ("visit_code", "visit_title")



from django.contrib.admin import SimpleListFilter
from django.contrib.sites.shortcuts import get_current_site

from ..models import Medication, Rx


class MedicationsListFilter(SimpleListFilter):

    title = "Medication"
    parameter_name = "medication_name"

    def lookups(self, request, model_admin):
        medications = []
        for medication in Medication.objects.all().order_by("name"):
            medications.append((medication.name, medication.name.replace("_", " ").title()))
        medications.append(("none", "None"))
        return tuple(medications)

    def queryset(self, request, queryset):
        """Returns a queryset if the Medication name is in the list of sites"""
        qs = None
        if self.value():
            if self.value() == "none":
                qs = Rx.objects.filter(
                    medications__isnull=True, site=get_current_site(request)
                )
            else:
                qs = Rx.objects.filter(
                    medications__name__in=[self.value()], site=get_current_site(request)
                )
        return qs



from django.apps import apps as django_apps

app_name = "edc_pharmacy"
# groups
DISPENSING = "DISPENSING"
DISPENSING_VIEW = "DISPENSING_VIEW"
PHARMACY = "PHARMACY"
PHARMACY_PRESCRIBER = "PHARMACY_PRESCRIBER"
PHARMACY_VIEW = "PHARMACY_VIEW"

# roles
CENTRAL_PHARMACIST_ROLE = "CENTRAL_PHARMACIST_ROLE"
PHARMACIST_ROLE = "PHARMACIST_ROLE"
PHARMACY_AUDITOR_ROLE = "PHARMACY_AUDITOR_ROLE"
PHARMACY_PRESCRIBER_ROLE = "PHARMACY_PRESCRIBER_ROLE"
SITE_PHARMACIST_ROLE = "SITE_PHARMACIST_ROLE"

pharmacy_codenames = ["edc_pharmacy.view_subject"]
prescriber_codenames = ["edc_pharmacy.view_subject"]

for app_config in django_apps.get_app_configs():
    if app_config.name in [
        "edc_pharmacy",
    ]:
        for model_cls in app_config.get_models():
            app_name, model_name = model_cls._meta.label_lower.split(".")
            if model_name == "subject":
                continue
            for prefix in ["add", "change", "view", "delete"]:
                pharmacy_codenames.append(f"{app_name}.{prefix}_{model_name}")

for model_name in ["dosageguideline", "formulation", "medication", "rxrefill"]:
    prescriber_codenames.extend(
        [
            c
            for c in pharmacy_codenames
            if model_name in c and c.startswith("edc_pharmacy.view")
        ]
    )
for model_name in ["rx", "rxitem"]:
    prescriber_codenames.extend([c for c in pharmacy_codenames if model_name in c])

prescriber_codenames.sort()
pharmacy_codenames.sort()



from django.db import models
from django.db.models.deletion import PROTECT
from edc_model import models as edc_models
from edc_utils import get_utcnow

from ..choices import DISPENSE_STATUS
from ..constants import DISPENSED
from ..dispense import Dispensing
from .rx_refill import RxRefill


class Manager(models.Manager):

    use_in_migrations = True

    def get_by_natural_key(self, rx_refill, dispensed_datetime):
        return self.get(rx_refill, dispensed_datetime)


class DispensingHistory(edc_models.BaseUuidModel):
    """A model to capture an amount dispensed against a refill"""

    rx_refill = models.ForeignKey(RxRefill, on_delete=PROTECT)

    dispensed_datetime = models.DateTimeField(default=get_utcnow)

    dispensed = models.DecimalField(max_digits=6, decimal_places=1)

    status = models.CharField(
        verbose_name="Status", max_length=25, default=DISPENSED, choices=DISPENSE_STATUS
    )

    objects = Manager()

    history = edc_models.HistoricalRecords()

    def __str__(self):
        return f"{str(self.rx_refill)}"

    def natural_key(self):
        return (
            self.rx_refill,
            self.dispensed_datetime,
        )

    def save(self, *args, **kwargs):
        # will raise an exception if nothing remaining on refill
        Dispensing(rx_refill=self.rx_refill, dispensed=self.dispensed, exclude_id=self.id)
        super().save(*args, **kwargs)

    @property
    def dispensed_date(self):
        return self.dispensed_datetime.date()

    class Meta(edc_models.BaseUuidModel.Meta):
        verbose_name = "Dispensing history"
        verbose_name_plural = "Dispensing history"
        unique_together = ["rx_refill", "dispensed_datetime"]



from edc_auth.site_auths import site_auths

from .auth_objects import (
    PHARMACIST_ROLE,
    PHARMACY,
    PHARMACY_AUDITOR_ROLE,
    PHARMACY_PRESCRIBER,
    PHARMACY_PRESCRIBER_ROLE,
    PHARMACY_VIEW,
    SITE_PHARMACIST_ROLE,
    pharmacy_codenames,
    prescriber_codenames,
)

site_auths.add_group(*pharmacy_codenames, name=PHARMACY_VIEW, view_only=True)
site_auths.add_group(*pharmacy_codenames, name=PHARMACY, no_delete=False)
site_auths.add_group(*prescriber_codenames, name=PHARMACY_PRESCRIBER, no_delete=True)

site_auths.add_role(PHARMACY, name=PHARMACIST_ROLE)
site_auths.add_role(PHARMACY, name=SITE_PHARMACIST_ROLE)
site_auths.add_role(PHARMACY_PRESCRIBER, name=PHARMACY_PRESCRIBER_ROLE)
site_auths.add_role(PHARMACY_VIEW, name=PHARMACY_AUDITOR_ROLE)



from typing import Tuple

from django.contrib import admin
from django.template.loader import render_to_string
from django.urls import reverse
from django_audit_fields.admin import audit_fieldset_tuple

from ..admin_site import edc_pharmacy_admin
from ..forms import ReturnHistoryForm
from ..models import ReturnHistory
from .model_admin_mixin import ModelAdminMixin


@admin.register(ReturnHistory, site=edc_pharmacy_admin)
class ReturnHistoryAdmin(ModelAdminMixin, admin.ModelAdmin):

    form = ReturnHistoryForm

    fieldsets = (
        (
            None,
            {
                "fields": (
                    "rx_refill",
                    "returned",
                    "return_datetime",
                )
            },
        ),
        audit_fieldset_tuple,
    )

    list_display: Tuple[str, ...] = (
        "subject_identifier",
        "refill",
        "rx_refill",
        "returned",
        "return_datetime",
    )

    list_filter: Tuple[str, ...] = ("return_datetime",)

    search_fields: Tuple[str, ...] = (
        "prescription_item__id",
        "prescription_item__prescription__subject_identifier",
        "prescription_item__medication__name",
    )

    ordering: Tuple[str, ...] = ("return_datetime",)

    @admin.display(description="Subject identifier")
    def subject_identifier(self, obj=None):
        return obj.rx_refill.rx.subject_identifier

    @admin.display(description="Item")
    def refill(self, obj=None):
        url = reverse("edc_pharmacy_admin:edc_pharmacy_rxrefill_changelist")
        url = f"{url}?q={obj.rx_refill.id}"
        context = dict(title="Back to refill", url=url, label="Refill")
        return render_to_string("dashboard_button.html", context=context)



from django.contrib import admin
from django_audit_fields.admin import audit_fieldset_tuple

from ..admin_site import edc_pharmacy_admin
from ..forms import StockReceivingForm
from ..models import StockReceiving
from .model_admin_mixin import ModelAdminMixin


@admin.register(StockReceiving, site=edc_pharmacy_admin)
class StockReceivingAdmin(ModelAdminMixin, admin.ModelAdmin):

    show_object_tools = True

    form = StockReceivingForm

    fieldsets = (
        (
            None,
            {
                "fields": (
                    "product",
                    "qty",
                    "stock_identifiers",
                    "received",
                    "received_datetime",
                )
            },
        ),
        audit_fieldset_tuple,
    )

    list_display = (
        "product",
        "qty",
        "received",
        "received_datetime",
        "created",
        "modified",
    )
    list_filter = (
        "product",
        "received",
        "received_datetime",
        "created",
        "modified",
    )
    search_fields = ("product__product_identifier", "product__lot_no")
    ordering = ("received_datetime",)
    readonly_fields = ("received", "received_datetime")



class PrescriptionError(Exception):
    pass


class PrescriptionNotStarted(Exception):
    pass


class PrescriptionExpired(Exception):
    pass


class PrescriptionAlreadyExists(Exception):
    pass


class ActivePrescriptionRefillOverlap(Exception):
    pass


class RefillError(Exception):
    pass


class RefillAlreadyExists(Exception):
    pass


class ActiveRefillAlreadyExists(Exception):
    pass


class NextRefillError(Exception):
    pass



from django.contrib import admin
from django_audit_fields.admin import audit_fieldset_tuple

from ..admin_site import edc_pharmacy_admin
from ..forms import StockForm
from ..models import Stock
from .model_admin_mixin import ModelAdminMixin


@admin.register(Stock, site=edc_pharmacy_admin)
class StockAdmin(ModelAdminMixin, admin.ModelAdmin):

    show_object_tools = True

    form = StockForm

    fieldsets = (
        (
            None,
            {"fields": ("stock_identifier", "product")},
        ),
        audit_fieldset_tuple,
    )

    list_display = (
        "stock_identifier",
        "product",
        "created",
        "modified",
    )
    list_filter = (
        "product",
        "product__lot_no",
        "product__formulation",
        "created",
        "modified",
    )
    search_fields = ("stock_identifier", "product__lot_no")
    ordering = ("stock_identifier",)
    readonly_fields = ("stock_identifier", "product")



from typing import Tuple

from django.contrib import admin
from django_audit_fields.admin import audit_fieldset_tuple

from ..admin_site import edc_pharmacy_admin
from ..forms import MedicationForm
from ..models import Medication
from .model_admin_mixin import ModelAdminMixin


@admin.register(Medication, site=edc_pharmacy_admin)
class MedicationAdmin(ModelAdminMixin, admin.ModelAdmin):

    show_object_tools = True

    form = MedicationForm

    fieldsets = (
        (
            None,
            {"fields": ("name", "display_name", "notes")},
        ),
        audit_fieldset_tuple,
    )

    list_display: Tuple[str, ...] = ("name", "display_name", "created", "modified")

    search_fields: Tuple[str, ...] = ("name",)

    ordering: Tuple[str, ...] = ("name",)



from uuid import uuid4

from django.contrib import admin
from django.http import HttpResponseRedirect
from django.urls import reverse
from edc_label import Label
from edc_protocol import Protocol


@admin.action(permissions=["view"], description="Print medication stock labels")
def print_stock_labels(modeladmin, request, queryset):
    zpl_data = []
    label = Label(
        label_template_name="stock_label.lbl",
        static_files_path="edc_pharmacy/label_templates",
    )
    for obj in queryset:
        for i in range(1, obj.qty):
            stock_identifier = uuid4().hex
            context = dict(
                protocol=Protocol().protocol,
                protocol_title=Protocol().protocol_title,
                stock_identifier=stock_identifier,
                barcode_value=stock_identifier,
                medication_name=obj.product.formulation.medication.display_name,
                medication_strength=obj.product.formulation.strength,
                medication_units=obj.product.formulation.units,
            )
            # keys = [k for k in context]
            # for fld in obj._meta.get_fields():
            #     if fld.name not in keys and fld.name != "assignment":
            #         context.update({fld.name: getattr(obj, fld.name)})
            zpl_data.append(
                str(label.render_as_zpl_data(copies=1, context=context, encoding=False))
                .strip("\n")
                .replace("\n", "")
            )
    request.session["zpl_data"] = "|".join(zpl_data)
    url = reverse("edc_label:browser_print_labels_url")
    return HttpResponseRedirect(url)



from uuid import uuid4

from django.db import models
from django.db.models import PROTECT
from edc_constants.choices import YES_NO
from edc_constants.constants import YES


class StudyMedicationRefillModelMixin(models.Model):

    refill_start_datetime = models.DateTimeField()

    refill_end_datetime = models.DateTimeField(
        null=True,
        blank=True,
        help_text="Leave blank to auto-calculate if refilling to the next scheduled visit",
    )

    refill_identifier = models.CharField(max_length=36, default=uuid4, editable=False)

    dosage_guideline = models.ForeignKey(
        "edc_pharmacy.DosageGuideline", on_delete=PROTECT, null=True, blank=False
    )

    formulation = models.ForeignKey(
        "edc_pharmacy.Formulation", on_delete=PROTECT, null=True, blank=False
    )

    roundup_divisible_by = models.IntegerField(default=1)

    refill_to_next_visit = models.CharField(
        verbose_name="Refill to the next scheduled visit",
        max_length=25,
        choices=YES_NO,
        default=YES,
        help_text="If YES, leave refill end date blank to auto-calculate",
    )

    number_of_days = models.IntegerField(
        null=True,
        blank=True,
        help_text="Leave blank to auto-calculate relative to the next scheduled appointment",
    )

    special_instructions = models.TextField(null=True, blank=True)

    order_or_update_next = models.CharField(
        verbose_name="Order, or update, refill for next scheduled visit?",
        max_length=15,
        choices=YES_NO,
        default=YES,
    )

    next_dosage_guideline = models.ForeignKey(
        "edc_pharmacy.DosageGuideline",
        on_delete=PROTECT,
        related_name="next_dosageguideline",
        null=True,
        blank=True,
    )

    next_formulation = models.ForeignKey(
        "edc_pharmacy.Formulation",
        on_delete=PROTECT,
        related_name="next_formulation",
        null=True,
        blank=True,
    )

    class Meta:
        verbose_name = "Study Medication"
        verbose_name_plural = "Study Medication"
        abstract = True



from edc_action_item import Action, site_action_items
from edc_constants.constants import HIGH_PRIORITY

from .constants import PRESCRIPTION_ACTION


class PrescriptionAction(Action):
    name = PRESCRIPTION_ACTION
    display_name = "New prescription"
    reference_model = "edc_pharmacy.rx"
    priority = HIGH_PRIORITY
    show_on_dashboard = True
    show_link_to_changelist = True
    admin_site_name = "edc_pharmacy_admin"
    create_by_user = False
    singleton = True
    instructions = (
        "This prescription will cover the participant for the period on study. "
        "All refills are written against this prescription."
    )

    def reopen_action_item_on_change(self):
        return False


site_action_items.register(PrescriptionAction)



from django.db import models
from django.db.models import PROTECT
from edc_model import models as edc_models

from .formulation import Formulation


class Manager(models.Manager):

    use_in_migrations = True


class MedicationLot(edc_models.BaseUuidModel):

    lot_no = models.CharField(max_length=50, unique=True)

    expiration_date = models.DateField()

    formulation = models.ForeignKey(Formulation, on_delete=PROTECT)

    objects = Manager()

    history = edc_models.HistoricalRecords()

    def __str__(self):
        return self.lot_no

    class Meta(edc_models.BaseUuidModel.Meta):
        verbose_name = "Medication lot"
        verbose_name_plural = "Medication lots"



from __future__ import annotations

from datetime import datetime
from typing import Any

from .refill_creator import RefillCreator


def create_refill(
    refill_identifier: str,
    subject_identifier: str,
    dosage_guideline: Any,
    formulation: Any,
    refill_start_datetime: datetime,
    refill_end_datetime: datetime,
    roundup_divisible_by: int | None,
    weight_in_kgs: float | None,
) -> RefillCreator:
    """Creates the edc_pharmacy refill for this study medication CRF,
    if not already created.

    Usually called by study medication CRF

    Called from signal.
    """
    rx_refill = RefillCreator(
        refill_identifier=refill_identifier,
        dosage_guideline=dosage_guideline,
        formulation=formulation,
        make_active=True,
        refill_start_datetime=refill_start_datetime,
        refill_end_datetime=refill_end_datetime,
        roundup_divisible_by=roundup_divisible_by,
        subject_identifier=subject_identifier,
        weight_in_kgs=weight_in_kgs,
    )
    # adjust_previous

    return rx_refill



from django.contrib import admin
from django_audit_fields.admin import audit_fieldset_tuple

from ..admin_site import edc_pharmacy_admin
from ..forms import ProductForm
from ..models import Product
from .model_admin_mixin import ModelAdminMixin


@admin.register(Product, site=edc_pharmacy_admin)
class ProductAdmin(ModelAdminMixin, admin.ModelAdmin):

    show_object_tools = True

    form = ProductForm

    fieldsets = (
        (
            None,
            {
                "fields": (
                    [
                        "product_identifier",
                        "container",
                        "count_per_container",
                        "formulation",
                        "lot_no",
                    ]
                )
            },
        ),
        audit_fieldset_tuple,
    )

    list_display = (
        "product_identifier",
        "container",
        "count_per_container",
        "formulation",
        "lot_no",
        "created",
        "modified",
    )
    list_filter = (
        "formulation",
        "container",
    )
    search_fields = (
        "product_identifier",
        "lot_no",
    )
    ordering = (
        "product_identifier",
        "lot_no",
    )
    readonly_fields = ("product_identifier",)



from django.contrib import admin
from edc_list_data.admin import ListModelAdminMixin

from ..admin_site import edc_pharmacy_admin
from ..models import Container, FormulationType, FrequencyUnits, Route, Units


@admin.register(Container, site=edc_pharmacy_admin)
class ContainerAdmin(ListModelAdminMixin, admin.ModelAdmin):
    pass


@admin.register(FormulationType, site=edc_pharmacy_admin)
class FormulationTypeAdmin(ListModelAdminMixin, admin.ModelAdmin):
    pass


@admin.register(FrequencyUnits, site=edc_pharmacy_admin)
class FrequencyUnitsAdmin(ListModelAdminMixin, admin.ModelAdmin):
    pass


@admin.register(Route, site=edc_pharmacy_admin)
class RouteAdmin(ListModelAdminMixin, admin.ModelAdmin):
    pass


@admin.register(Units, site=edc_pharmacy_admin)
class UnitsAdmin(ListModelAdminMixin, admin.ModelAdmin):
    pass



from django.contrib import admin
from django.template.loader import render_to_string
from django.urls import reverse
from django_audit_fields.admin import audit_fieldset_tuple
from edc_model_admin.dashboard import ModelAdminSubjectDashboardMixin

from ..admin_site import edc_pharmacy_admin
from ..forms import RxForm
from ..models import Rx
from .list_filters import MedicationsListFilter


@admin.register(Rx, site=edc_pharmacy_admin)
class RxAdmin(ModelAdminSubjectDashboardMixin, admin.ModelAdmin):

    show_object_tools = True

    form = RxForm

    fieldsets = (
        (
            None,
            {
                "fields": (
                    "subject_identifier",
                    "report_datetime",
                    "rx_name",
                    "rx_date",
                    "medications",
                    "clinician_initials",
                    "notes",
                )
            },
        ),
        (
            "Randomization",
            {
                "fields": ("rando_sid", "randomizer_name", "weight_in_kgs"),
            },
        ),
        audit_fieldset_tuple,
    )

    filter_horizontal = ("medications",)

    list_display = (
        "subject_identifier",
        "dashboard",
        "add_refill",
        "refills",
        "rx_medications",
        "rando_sid",
        "rx_date",
        "weight_in_kgs",
        "rx_name",
    )

    list_filter = ("report_datetime", MedicationsListFilter, "site")

    search_fields = (
        "id",
        "subject_identifier",
        "rando_sid",
        "registered_subject__initials",
        "medications__name",
        "site__id",
        "rx_name",
    )

    readonly_fields = (
        "rando_sid",
        "weight_in_kgs",
        "rx_name",
    )

    @admin.display
    def add_refill(self, obj=None, label=None):
        url = reverse("edc_pharmacy_admin:edc_pharmacy_rxrefill_add")
        url = f"{url}?rx={obj.id}"
        context = dict(title="Add refill", url=url, label="Add refill")
        return render_to_string("dashboard_button.html", context=context)

    @admin.display
    def refills(self, obj=None, label=None):
        url = reverse("edc_pharmacy_admin:edc_pharmacy_rxrefill_changelist")
        url = f"{url}?q={obj.id}"
        context = dict(title="RX items", url=url, label="Refills")
        return render_to_string("dashboard_button.html", context=context)

    @admin.display
    def rx_medications(self, obj):
        return ", ".join([obj.display_name for obj in obj.medications.all()])



from django.test import TestCase
from edc_list_data import site_list_data

from edc_pharmacy.dosage_calculator import DosageCalculator
from edc_pharmacy.dosage_per_day import DosageError
from edc_pharmacy.models import (
    DosageGuideline,
    Formulation,
    FormulationType,
    FrequencyUnits,
    Medication,
    Route,
    Units,
)


class TestDoseCalculator(TestCase):
    def setUp(self):
        site_list_data.initialize()
        site_list_data.autodiscover()
        medication = Medication.objects.create(
            name="flucytosine",
            display_name="Flucytosine",
        )

        Formulation.objects.create(
            medication=medication,
            strength=500,
            units=Units.objects.get(display_name__iexact="mg"),
            route=Route.objects.get(display_name__iexact="oral"),
            formulation_type=FormulationType.objects.all()[0],
        )

        medication = Medication.objects.create(
            name="flucanazole",
            display_name="Flucanazole",
        )

        Formulation.objects.create(
            medication=medication,
            strength=200,
            units=Units.objects.get(display_name__iexact="mg"),
            route=Route.objects.get(display_name__iexact="oral"),
            formulation_type=FormulationType.objects.all()[0],
        )

        medication = Medication.objects.create(
            name="ambisome",
            display_name="Ambisome",
        )

        Formulation.objects.create(
            medication=medication,
            strength=50,
            units=Units.objects.get(display_name__iexact="mg"),
            route=Route.objects.get(display_name__iexact="intravenous"),
            formulation_type=FormulationType.objects.all()[0],
        )

    def test_dosage_flucytosine(self):
        medication = Medication.objects.get(name="flucytosine")
        dosage_guideline = DosageGuideline.objects.create(
            medication=medication,
            dose_per_kg=100.0,
            dose_units=Units.objects.get(display_name__iexact="mg"),
            frequency=1.0,
            frequency_units=FrequencyUnits.objects.get(display_name__iexact="times per day"),
        )
        formulation = Formulation.objects.get(
            medication=medication,
        )

        self.assertEqual(
            DosageCalculator(
                dosage_guideline=dosage_guideline,
                formulation=formulation,
                weight_in_kgs=100.0,
            ).dosage,
            20.0,
        )
        self.assertEqual(
            DosageCalculator(
                dosage_guideline=dosage_guideline,
                formulation=formulation,
                weight_in_kgs=40.0,
            ).dosage,
            8.0,
        )

    def test_dosage_ambisome(self):
        medication = Medication.objects.get(name="ambisome")
        dosage_guideline = DosageGuideline.objects.create(
            medication=medication,
            dose_per_kg=10.0,
            dose_units=Units.objects.get(display_name__iexact="mg"),
            frequency=1.0,
            frequency_units=FrequencyUnits.objects.get(display_name__iexact="times per day"),
        )
        formulation = Formulation.objects.get(medication=medication)
        self.assertEqual(
            DosageCalculator(
                dosage_guideline=dosage_guideline, formulation=formulation, weight_in_kgs=40.0
            ).dosage,
            8.0,
        )

    def test_dosage_flucanazole(self):
        medication = Medication.objects.get(name="flucanazole")
        dosage_guideline = DosageGuideline.objects.create(
            medication=medication,
            dose=1200.0,
            dose_units=Units.objects.get(display_name__iexact="mg"),
            frequency=1.0,
            frequency_units=FrequencyUnits.objects.get(display_name__iexact="times per day"),
        )
        formulation = Formulation.objects.get(medication=medication)
        self.assertEqual(
            DosageCalculator(
                dosage_guideline=dosage_guideline, formulation=formulation
            ).dosage,
            6.0,
        )

    def test_dosage_exceptions(self):
        medication = Medication.objects.get(name="flucytosine")
        dosage_guideline = DosageGuideline.objects.create(
            medication=medication,
            dose_per_kg=100,
            dose_units=Units.objects.get(display_name__iexact="mg"),
            frequency=1,
            frequency_units=FrequencyUnits.objects.get(display_name__iexact="times per day"),
        )

        formulation = Formulation.objects.create(
            medication=medication,
            strength=200,
            units=Units.objects.get(display_name__iexact="g"),
            route=Route.objects.get(display_name__iexact="oral"),
            formulation_type=FormulationType.objects.all()[0],
        )

        self.assertRaises(
            DosageError,
            DosageCalculator,
            dosage_guideline=dosage_guideline,
            formulation=formulation,
            weight_in_kgs=40,
        )



from django.db.models import Sum


class DispenseError(Exception):
    pass


# TODO: dispense against stock for site / central, etc
class Dispensing:
    def __init__(
        self,
        rx_refill,
        dispensed=None,
        exclude_id=None,
    ):
        """Dispense against an existing refill with remaining items"""
        self.exclude_id = exclude_id
        self.rx_refill = rx_refill
        self.dispensed = dispensed

    @property
    def remaining(self) -> float:
        value = 0.0
        if self.rx_refill.total:
            value = float(self.rx_refill.total) - float(self.total_dispensed)
        if value < 0.0:
            raise DispenseError(
                "Attempt to dispense more than remaining on refill. "
                f"Remaining={value}. Got {self.dispensed}."
            )
        return value

    @property
    def total_dispensed(self) -> float:
        """Returns the total dispensed for this refill"""
        options = {}
        if self.rx_refill.total:
            if self.exclude_id:
                options = dict(id=self.exclude_id)
            aggregate = self.rx_refill.dispensinghistory_set.filter(**options).aggregate(
                Sum("dispensed")
            )
            return float(aggregate.get("dispensed__sum") or 0.0)
        return 0.0



from django.test import TestCase
from edc_list_data import site_list_data

from edc_pharmacy.dosage_calculator import DosageCalculator
from edc_pharmacy.models import (
    DosageGuideline,
    Formulation,
    FormulationType,
    FrequencyUnits,
    Medication,
    Route,
    Units,
)


class TestDoseCalculator(TestCase):
    def setUp(self):
        site_list_data.initialize()
        site_list_data.autodiscover()
        self.medication = Medication.objects.create(
            name="Metformin",
            display_name="Metformin",
        )

        formulation_type = FormulationType.objects.get(name="tablet")

        self.formulation = Formulation.objects.create(
            medication=self.medication,
            strength=500,
            units=Units.objects.get(display_name__iexact="mg"),
            route=Route.objects.get(display_name__iexact="oral"),
            formulation_type=formulation_type,
        )
        self.dosage_guideline_baseline = DosageGuideline.objects.create(
            medication=self.medication,
            dose=1000.0,
            dose_units=Units.objects.get(display_name__iexact="mg"),
            frequency=1.0,
            frequency_units=FrequencyUnits.objects.get(display_name__iexact="times per day"),
        )
        self.dosage_guideline = DosageGuideline.objects.create(
            medication=self.medication,
            dose=2000.0,
            dose_units=Units.objects.get(display_name__iexact="mg"),
            frequency=1.0,
            frequency_units=FrequencyUnits.objects.get(display_name__iexact="times per day"),
        )

    def test_dosage_baseline(self):
        """2 tabs / day = 1000mg"""
        self.assertEqual(
            DosageCalculator(
                dosage_guideline=self.dosage_guideline_baseline, formulation=self.formulation
            ).dosage,
            2.0,
        )

    def test_dosage_post_baseline(self):
        """4 tabs / day = 2000mg"""
        self.assertEqual(
            DosageCalculator(
                dosage_guideline=self.dosage_guideline, formulation=self.formulation
            ).dosage,
            4.0,
        )



from django.conf import settings
from django.contrib import admin
from django_audit_fields.admin import audit_fieldset_tuple
from edc_utils import convert_php_dateformat

from ..admin_site import edc_pharmacy_admin
from ..forms import MedicationOrderForm
from ..models import MedicationOrder
from .model_admin_mixin import ModelAdminMixin


@admin.register(MedicationOrder, site=edc_pharmacy_admin)
class MedicationOrderAdmin(ModelAdminMixin, admin.ModelAdmin):

    show_object_tools = True

    autocomplete_fields = (
        "dosage_guideline",
        "formulation",
    )

    form = MedicationOrderForm

    model = MedicationOrder

    fieldsets = (
        (
            "Order",
            {
                "fields": (
                    "stock",
                    "qty",
                ),
            },
        ),
        (
            "Order status",
            {
                "fields": (
                    "packed",
                    "packed_datetime",
                    "shipped",
                    "shipped_datetime",
                    "received_at_site",
                    "received_at_site_datetime",
                ),
            },
        ),
        (
            "Refill",
            {
                "fields": (
                    "rx",
                    "dosage_guideline",
                    "formulation",
                    "refill_date",
                    "number_of_days",
                )
            },
        ),
        audit_fieldset_tuple,
    )

    list_display = (
        "subject_identifier",
        "refill_date",
        "description",
        "consented",
        "packed",
        "shipped",
        "received_at_site",
    )
    list_filter = (
        "packed",
        "shipped",
        "received_at_site",
        "refill_date",
        "visit_code",
        "visit_code_sequence",
        "site",
    )
    search_fields = (
        "id",
        "site__id",
        "rx__id",
        "rx__rando_sid",
        "rx__subject_identifier",
        "rx__registered_subject__initials",
        "dosage_guideline__medication__name",
    )
    ordering = (
        "rx__subject_identifier",
        "-refill_date",
    )

    readonly_fields = (
        "rx",
        "dosage_guideline",
        "formulation",
        "refill_date",
        "number_of_days",
        "packed",
        "shipped",
        "received_at_site",
    )

    @staticmethod
    def description(obj):
        dob = obj.rx.registered_subject.dob.strftime(
            convert_php_dateformat(settings.SHORT_DATE_FORMAT)
        )
        return f"{obj.rx.registered_subject.initials} {dob} {obj.rx.registered_subject.gender}"

    @staticmethod
    def gender(obj):
        return obj.rx.registered_subject.gender

    @staticmethod
    def consented(obj):
        return obj.rx.registered_subject.consent_datetime.strftime(
            convert_php_dateformat(settings.SHORT_DATE_FORMAT)
        )

    @admin.display(description="Subject")
    def subject_identifier(self, obj=None):
        return obj.rx.subject_identifier



from django import template
from django.conf import settings
from edc_dashboard.utils import get_bootstrap_version

register = template.Library()


@register.inclusion_tag(
    f"edc_pharmacy/bootstrap{get_bootstrap_version()}/prescription_item_description.html",
    takes_context=True,
)
def format_prescription_description(context, prescription_item):
    context["SHORT_DATE_FORMAT"] = settings.SHORT_DATE_FORMAT
    context["prescription_item"] = prescription_item
    return context



from .autocomplete_admin import SubjectAdmin
from .dispensing_history_admin import DispensingHistoryAdmin
from .dosage_guideline_admin import DosageGuidelineAdmin
from .formulation_admin import FormulationAdmin
from .labels_admin import LabelsAdmin
from .list_model_admin import (
    FormulationTypeAdmin,
    FrequencyUnitsAdmin,
    RouteAdmin,
    UnitsAdmin,
)
from .medication_admin import MedicationAdmin
from .medication_lot_admin import MedicationLotAdmin
from .product_admin import ProductAdmin
from .return_history_admin import ReturnHistoryAdmin
from .rx_admin import RxAdmin
from .rx_refill_admin import RxRefillAdmin
from .stock_admin import StockAdmin
from .stock_create_labels_admin import StockCreateLabelsAdmin
from .stock_receiving_admin import StockReceivingAdmin



from __future__ import annotations

from edc_appointment.utils import get_next_appointment

from ..exceptions import NextRefillError
from ..refill import create_refills_from_crf
from .previous_model_mixin import PreviousNextModelMixin
from .study_medication_refill_model_mixin import StudyMedicationRefillModelMixin


class StudyMedicationCrfModelMixin(PreviousNextModelMixin, StudyMedicationRefillModelMixin):

    """Declare with field subject_visit using a CRF model mixin"""

    def save(self, *args, **kwargs):
        if not get_next_appointment(self.related_visit.appointment, include_interim=True):
            raise NextRefillError(
                "Cannot refill. This subject has no future appointments. "
                f"See {self.related_visit}."
            )
        if not self.refill_end_datetime:
            self.refill_end_datetime = get_next_appointment(
                self.related_visit.appointment, include_interim=True
            ).appt_datetime
        self.adjust_end_datetimes()
        self.number_of_days = (self.refill_end_datetime - self.refill_start_datetime).days
        super().save(*args, **kwargs)

    def creates_refills_from_crf(self) -> tuple:
        """Attribute called in signal"""
        return create_refills_from_crf(self, self.related_visit_model_attr())

    def get_subject_identifier(self):
        return self.related_visit.subject_identifier

    class Meta(StudyMedicationRefillModelMixin.Meta):
        abstract = True



from ...utils import get_rxrefill_model_cls


def delete_next_refills_for_crf(instance):
    if (
        qs := get_rxrefill_model_cls()
        .objects.filter(
            rx__subject_identifier=instance.subject_visit.subject_identifier,
            refill_start_datetime__gt=instance.refill_start_datetime,
            active=False,
        )
        .order_by("refill_start_datetime")
    ):
        qs[0].delete()



from django.contrib import admin
from django_audit_fields.admin import audit_fieldset_tuple

from ..admin_site import edc_pharmacy_admin
from ..forms import StockCreateLabelsForm
from ..models import StockCreateLabels
from .actions import print_stock_labels
from .model_admin_mixin import ModelAdminMixin


@admin.register(StockCreateLabels, site=edc_pharmacy_admin)
class StockCreateLabelsAdmin(ModelAdminMixin, admin.ModelAdmin):

    show_object_tools = True

    form = StockCreateLabelsForm

    actions = [print_stock_labels]

    fieldsets = (
        (
            None,
            {"fields": ("product", "qty", "printed", "printed_datetime")},
        ),
        audit_fieldset_tuple,
    )

    list_display = (
        "product",
        "qty",
        "printed",
        "printed_datetime",
        "created",
        "modified",
    )
    list_filter = (
        "product",
        "printed",
        "printed_datetime",
        "created",
        "modified",
    )
    search_fields = (
        "product__product_identifier",
        "product__lot_no",
    )
    ordering = ("printed_datetime",)
    readonly_fields = ("printed", "printed_datetime")



from uuid import uuid4

from django.core.exceptions import ObjectDoesNotExist
from django.db import models
from django.db.models import PROTECT
from edc_action_item.models import ActionModelMixin
from edc_constants.constants import NEW
from edc_identifier.model_mixins import NonUniqueSubjectIdentifierFieldMixin
from edc_model import models as edc_models
from edc_randomization.site_randomizers import site_randomizers
from edc_registration.models import RegisteredSubject
from edc_sites.models import CurrentSiteManager, SiteModelMixin
from edc_utils import formatted_age, get_utcnow

from ..choices import PRESCRIPTION_STATUS
from ..constants import PRESCRIPTION_ACTION
from ..exceptions import PrescriptionError
from .medication import Medication
from .search_slug_model_mixin import SearchSlugModelMixin
from .subject import Subject


class Rx(
    NonUniqueSubjectIdentifierFieldMixin,
    SiteModelMixin,
    ActionModelMixin,
    SearchSlugModelMixin,
    edc_models.BaseUuidModel,
):

    """A model for the prescription.

    In this context the `prescription` specifies only the medication.
    The formulation and doseage guidelines are specified
    on each Refill.
    """

    action_name = PRESCRIPTION_ACTION

    rx_identifier = models.CharField(max_length=36, default=uuid4)

    rx_name = models.CharField(max_length=36, default="study prescription")

    registered_subject = models.ForeignKey(
        Subject,
        verbose_name="Subject Identifier",
        on_delete=PROTECT,
        null=True,
        blank=False,
    )

    report_datetime = models.DateTimeField(default=get_utcnow)

    rx_date = models.DateField(verbose_name="Date RX written", default=get_utcnow)

    rx_expiration_date = models.DateField(
        verbose_name="Date RX expires",
        null=True,
        blank=True,
        help_text="Leave blank. Will be filled when end of study report is submitted",
    )

    status = models.CharField(max_length=25, default=NEW, choices=PRESCRIPTION_STATUS)

    medications = models.ManyToManyField(Medication, blank=False)

    refill = models.IntegerField(
        null=True,
        blank=True,
        help_text="Number of times this prescription may be refilled",
    )

    rando_sid = models.CharField(max_length=25, null=True, blank=True)

    randomizer_name = models.CharField(max_length=25, null=True, blank=True)

    weight_in_kgs = models.DecimalField(max_digits=6, decimal_places=1, null=True, blank=True)

    clinician_initials = models.CharField(max_length=3, null=True)

    notes = models.TextField(
        max_length=250,
        null=True,
        blank=True,
        help_text="Private notes for pharmacist only",
    )

    on_site = CurrentSiteManager()

    def __str__(self):
        return f"{self.subject_identifier}"

    def natural_key(self):
        return (self.rx_identifier,)  # noqa

    def save(self, *args, **kwargs):
        self.registered_subject = RegisteredSubject.objects.get(
            subject_identifier=self.subject_identifier
        )
        if self.randomizer_name:
            randomizer = site_randomizers.get(self.randomizer_name)
            try:
                self.rando_sid = (
                    randomizer.model_cls()
                    .objects.get(subject_identifier=self.subject_identifier)
                    .sid
                )
            except ObjectDoesNotExist:
                raise PrescriptionError(
                    "Unable to create prescription. Subject has not been "
                    f"randomized (randomizer={self.randomizer_name}"
                )
        super().save(*args, **kwargs)

    def description(self):
        return (
            f"{','.join([o.display_name for o in self.medications.all()])} "
            f"{self.registered_subject.subject_identifier} {self.registered_subject.initials} "
            f"{formatted_age(born=self.registered_subject.dob, reference_dt=get_utcnow())} "
            f"{self.registered_subject.gender} "
            f"Written: {self.rx_date}"
        )

    class Meta(edc_models.BaseUuidModel.Meta):
        verbose_name = "Prescription"
        verbose_name_plural = "Prescriptions"



from __future__ import annotations

from copy import deepcopy
from datetime import datetime
from decimal import Decimal
from typing import Any
from uuid import uuid4
from zoneinfo import ZoneInfo

from dateutil.relativedelta import relativedelta
from django.conf import settings
from django.core.exceptions import ObjectDoesNotExist
from django.db import models
from edc_utils import convert_php_dateformat

from ..exceptions import PrescriptionError, PrescriptionExpired, PrescriptionNotStarted
from ..utils import get_rx_model_cls, get_rxrefill_model_cls
from .activate_refill import activate_refill


class RefillCreatorError(Exception):
    pass


class RefillCreator:
    def __init__(
        self,
        refill_identifier: str = None,
        subject_identifier: str | None = None,
        refill_start_datetime: datetime | None = None,
        refill_end_datetime: datetime | None = None,
        formulation: Any | None = None,
        dosage_guideline: models.Model | None = None,
        make_active: bool | None = None,
        force_active: bool | None = None,
        roundup_divisible_by: int | None = None,
        weight_in_kgs: float | Decimal | None = None,
    ):
        """Creates rx_refill"""
        self._next_rx_refill = None
        self._prev_rx_refill = None
        self.refill_identifier = refill_identifier
        self.refill_end_datetime = refill_end_datetime.astimezone(ZoneInfo("UTC"))
        self.subject_identifier = subject_identifier
        self.refill_start_datetime = refill_start_datetime.astimezone(ZoneInfo("UTC"))
        self.formulation = formulation
        self.dosage_guideline = dosage_guideline
        self.roundup_divisible_by = roundup_divisible_by or 0
        self.weight_in_kgs = weight_in_kgs

        if not self.weight_in_kgs:
            self.weight_in_kgs = self.rx.weight_in_kgs
        if self.dosage_guideline.dose_per_kg and not self.weight_in_kgs:
            raise RefillCreatorError("Dosage guideline requires patient's weight in kgs")

        self.make_active = True if make_active is None else make_active
        self.force_active = force_active
        self.rx_refill = self.create_or_update()
        if self.make_active:
            activate_refill(self.rx_refill)

    @property
    def options(self) -> dict:
        return dict(
            refill_identifier=self.refill_identifier or uuid4(),
            dosage_guideline=self.dosage_guideline,
            formulation=self.formulation,
            refill_start_datetime=self.refill_start_datetime,
            refill_end_datetime=self.refill_end_datetime,
            weight_in_kgs=self.weight_in_kgs,
            roundup_divisible_by=self.roundup_divisible_by,
        )

    def create_or_update(self) -> Any:
        """Creates / updates and returns a RxRefill."""
        # find first refill on or after this start date
        try:
            rx_refill = get_rxrefill_model_cls().objects.get(
                rx=self.rx, refill_start_datetime=self.refill_start_datetime
            )
        except ObjectDoesNotExist:
            opts = deepcopy(self.options)
            if self.prev_rx_refill:
                # found previous rx_refill, update end datetime, number_of_days
                self.prev_rx_refill.refill_end_datetime = (
                    self.refill_start_datetime - relativedelta(minutes=1)
                )
                self.prev_rx_refill.save()
            if self.next_rx_refill:
                refill_end_datetime = (
                    self.next_rx_refill.refill_start_datetime - relativedelta(minutes=1)
                )
                opts.update(refill_end_datetime=refill_end_datetime)
            # create a new rx_refill
            rx_refill = get_rxrefill_model_cls().objects.create(rx=self.rx, **opts)
        else:
            opts = deepcopy(self.options)
            refill_end_datetime = self.refill_end_datetime
            if self.next_rx_refill:
                refill_end_datetime = (
                    self.next_rx_refill.refill_start_datetime - relativedelta(minutes=1)
                )
            opts.update(refill_end_datetime=refill_end_datetime)
            for k, v in self.options.items():
                setattr(rx_refill, k, v)
            rx_refill.save()
            rx_refill.refresh_from_db()
        return rx_refill

    @property
    def rx(self) -> Any:
        """Returns Rx model instance else raises PrescriptionError"""
        opts = dict(
            subject_identifier=self.subject_identifier,
            medications__in=[self.formulation.medication],
        )
        try:
            rx = get_rx_model_cls().objects.get(**opts)
        except ObjectDoesNotExist:
            raise PrescriptionError(f"Subject does not have a prescription. Got {opts}.")
        else:
            refill_start_datetime = self.refill_start_datetime.strftime(
                convert_php_dateformat(settings.DATETIME_FORMAT)
            )
            if self.refill_start_datetime.date() < rx.rx_date:
                rx_date = rx.rx_date.strftime(convert_php_dateformat(settings.DATE_FORMAT))
                raise PrescriptionNotStarted(
                    f"Subject's prescription not started. Starts on {rx_date}. "
                    f"Got {self.subject_identifier} attempting "
                    f"refill on {refill_start_datetime}."
                )
            elif (
                rx.rx_expiration_date
                and self.refill_start_datetime.date() > rx.rx_expiration_date
            ):
                rx_expiration_date = rx.rx_expiration_date.strftime(
                    convert_php_dateformat(settings.DATE_FORMAT)
                )
                raise PrescriptionExpired(
                    f"Subject prescription has expired. Expired on {rx_expiration_date}. "
                    f"Got {self.subject_identifier} attempting refill "
                    f"on {refill_start_datetime}."
                )
        return rx

    @property
    def prev_rx_refill(self):
        if not self._prev_rx_refill:
            self._prev_rx_refill = (
                get_rxrefill_model_cls()
                .objects.filter(
                    rx=self.rx, refill_start_datetime__lt=self.refill_start_datetime
                )
                .order_by("refill_start_datetime")
                .first()
            )
        return self._prev_rx_refill

    @property
    def next_rx_refill(self):
        if not self._next_rx_refill:
            self._next_rx_refill = (
                get_rxrefill_model_cls()
                .objects.filter(
                    rx=self.rx, refill_start_datetime__gt=self.refill_start_datetime
                )
                .order_by("refill_start_datetime")
                .first()
            )
        return self._next_rx_refill



from typing import Any

from edc_constants.constants import YES

from ..create_next_refill import create_next_refill
from ..create_refill import create_refill
from .delete_next_refills_for_crf import delete_next_refills_for_crf


def calculate_days_to_next_refill(refill) -> int:
    """Returns the number of days until medication runs out"""
    return 0


def create_refills_from_crf(instance: Any, visit_model_attr: str) -> tuple:
    next_rx_refill = None
    subject_visit = getattr(instance, visit_model_attr)
    refill_creator = create_refill(
        refill_identifier=instance.refill_identifier,
        subject_identifier=subject_visit.subject_identifier,
        dosage_guideline=instance.dosage_guideline,
        formulation=instance.formulation,
        refill_start_datetime=instance.refill_start_datetime,
        refill_end_datetime=instance.refill_end_datetime,
        roundup_divisible_by=instance.roundup_divisible_by,
        weight_in_kgs=getattr(instance, "weight_in_kgs", None),
    )
    rx_refill = refill_creator.rx_refill
    if instance.order_or_update_next == YES:
        next_rx_refill = create_next_refill(instance, visit_model_attr)
    elif not getattr(instance, visit_model_attr).appointment.next:
        delete_next_refills_for_crf(instance)
    return rx_refill, next_rx_refill



from django.db import models
from django.db.models import PROTECT
from edc_model import models as edc_models
from edc_sites.models import SiteModelMixin


class Manager(models.Manager):

    use_in_migrations = True


class ModelMixin(SiteModelMixin, models.Model):

    name = models.CharField(max_length=25, unique=True)

    description = models.CharField(max_length=250)

    objects = Manager()

    history = edc_models.HistoricalRecords()

    def __str__(self):
        return self.name

    class Meta(SiteModelMixin.Meta):
        abstract = True


class Warehouse(ModelMixin, edc_models.BaseUuidModel):
    class Meta(ModelMixin.Meta, edc_models.BaseUuidModel.Meta):
        verbose_name = "Warehouse"
        verbose_name_plural = "Warehouses"


class Room(ModelMixin, edc_models.BaseUuidModel):

    warehouse = models.ForeignKey(Warehouse, on_delete=PROTECT)

    class Meta(ModelMixin.Meta, edc_models.BaseUuidModel.Meta):
        verbose_name = "Room"
        verbose_name_plural = "Rooms"


class Shelf(ModelMixin, edc_models.BaseUuidModel):

    room = models.ForeignKey(Room, on_delete=PROTECT)

    class Meta(ModelMixin.Meta, edc_models.BaseUuidModel.Meta):
        verbose_name = "Shelf"
        verbose_name_plural = "Shelves"


class Box(ModelMixin, edc_models.BaseUuidModel):

    shelf = models.ForeignKey(Shelf, on_delete=PROTECT)

    def __str__(self):
        return self.name

    class Meta(ModelMixin.Meta, edc_models.BaseUuidModel.Meta):
        verbose_name = "Box"
        verbose_name_plural = "Boxes"



from django.core.exceptions import ObjectDoesNotExist
from edc_appointment.utils import get_next_appointment
from edc_constants.constants import NO, YES
from edc_form_validators import INVALID_ERROR, FormValidator

from ...utils import get_rx_model_cls, get_rxrefill_model_cls


class StudyMedicationFormValidator(FormValidator):
    def clean(self):
        self.required_if(
            NO,
            field="refill_to_next_visit",
            field_required="refill_end_datetime",
            inverse=False,
        )
        if (
            self.cleaned_data.get("refill_start_datetime")
            and self.cleaned_data.get("refill_end_datetime")
            and self.cleaned_data.get("refill_start_datetime")
            >= self.cleaned_data.get("refill_end_datetime")
        ):
            self.raise_validation_error(
                {"refill_end_datetime": "Invalid. End date must be after the start date"},
                INVALID_ERROR,
            )
        self.required_if(
            YES, field="order_or_update_next", field_required="next_dosage_guideline"
        )
        if self.cleaned_data.get("order_or_update_next") == NO and self.next_refill:
            if self.next_refill.active:
                self.raise_validation_error(
                    "Invalid. Next refill is already active", INVALID_ERROR
                )
        if self.cleaned_data.get("order_or_update_next") == NO and not get_next_appointment(
            self.cleaned_data.get("subject_visit").appointment, include_interim=True
        ):
            self.raise_validation_error(
                "Invalid. This is the last scheduled visit", INVALID_ERROR
            )

        self.required_if(YES, field="order_or_update_next", field_required="next_formulation")

    @property
    def next_refill(self):
        for obj in (
            get_rxrefill_model_cls()
            .objects.filter(
                rx=self.rx,
                refill_start_datetime__gt=self.cleaned_data.get("refill_start_datetime"),
            )
            .order_by("refill_start_datetime")
        ):
            return obj
        return None

    @property
    def rx(self):
        try:
            return get_rx_model_cls().objects.get(
                subject_identifier=self.cleaned_data.get("subject_visit").subject_identifier,
                medications__in=[self.cleaned_data.get("formulation").medication],
            )
        except ObjectDoesNotExist:
            self.raise_validation_error(
                {"__all__": "Prescription does not exist"}, INVALID_ERROR
            )
